#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>
#include <CL/opencl.h>

#include <GL/glew.h>
#include <GL/freeglut.h>

#include "camera.hpp"
#include "matrix4x4.hpp"


//*********************************//
//       OpenCL inicializacio      //
//*********************************//
#pragma region R_CL_INIT

#define CL_SAFE_CALL( call ) {						\
    cl_int err = call;							\
    if(CL_SUCCESS != err){						\
      std::cout << "Unsuccesful OpenCL call " << __FILE__ << " : " << __LINE__ << std::endl; \
      exit(EXIT_FAILURE);						\
    } }

// OpenCL platform
cl_platform_id platform;

char* getPlatformInfo(cl_platform_id platform, cl_platform_info paramName){
  size_t infoSize = 0;
  CL_SAFE_CALL( clGetPlatformInfo(platform, paramName, 0, NULL, &infoSize) );
  char* info = (char*)malloc(infoSize);
  CL_SAFE_CALL( clGetPlatformInfo(platform, paramName, infoSize, info, NULL) );
  return info;
}

cl_platform_id createPlatform(){
  cl_platform_id platform;
  CL_SAFE_CALL( clGetPlatformIDs(1, &platform, NULL));
  std::cout << getPlatformInfo(platform, CL_PLATFORM_VERSION) << std::endl;
  return platform;
}

// OpenCL device-ok
cl_device_id device_id;

void* getDeviceInfo(cl_device_id device_id, cl_device_info paramName){
  size_t infoSize = 0;
  CL_SAFE_CALL( clGetDeviceInfo(device_id, paramName, 0, NULL, &infoSize) );
  char* info = (char*)malloc(infoSize);
  CL_SAFE_CALL( clGetDeviceInfo(device_id, paramName, infoSize, info, NULL) );
  return info;
}

cl_device_id createDevice(cl_platform_id platform, cl_device_type type){
  cl_device_id device_id;
  CL_SAFE_CALL( clGetDeviceIDs(platform, type, 1, &device_id, NULL) );

  cl_uint* max_compute_units = (cl_uint*)getDeviceInfo(device_id, CL_DEVICE_MAX_COMPUTE_UNITS);
  std::cout << "Max compute units: " << *max_compute_units << std::endl;

  return device_id;
}

// OpenCL kontextus
cl_context context;
cl_context createContext(cl_device_id device_id){
  cl_context context = 0;
  context = clCreateContext(0, 1, &device_id, NULL, NULL, NULL);
  if(!context){
    std::cerr << "Context creation failed!\n";
    exit(EXIT_FAILURE);
  }

  return context;
}

// OpenCL utasitassor
cl_command_queue commands;
cl_command_queue createCommandQueue(cl_context context, cl_device_id device){
  cl_command_queue command_queue = 0;
  command_queue = clCreateCommandQueue(context, device_id, 0, NULL);
  if(!command_queue){
    std::cerr << "Command queue creation failed!\n";
  }
  return command_queue;
}

// OpenCL program
cl_program program;

bool fileToString(const char* path, char*& out, int& len) {
  std::ifstream file(path, std::ios::ate | std::ios::binary);
  if(!file.is_open()) {
    return false;
  }
  len = file.tellg();
  out = new char[ len+1 ];
  file.seekg (0, std::ios::beg);
  file.read(out, len);
  file.close();
  out[len] = 0;
  return true;
}

cl_program createProgram(cl_context context, cl_device_id device_id, const char* fileName){
  char* programSource = NULL;
  int len = 0;
  int errorFlag = -1;
  if(!fileToString(fileName, programSource, len)){
    std::cerr << "Error loading program: " << fileName << std::endl;
    exit(EXIT_FAILURE);
  }
  cl_program program = 0;
  program = clCreateProgramWithSource(context, 1, (const char**)&programSource, NULL, NULL);
  if (!program) {
    std::cerr << "Error: Failed to create compute program!" << std::endl;
    exit(EXIT_FAILURE);
  }

  cl_int err = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
  if (err != CL_SUCCESS) {
    std::cerr << "Error: Failed to build program executable!" << std::endl;

    char buffer[10240];
    clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG,
			  sizeof(buffer), buffer, NULL);
    std::cerr << buffer << std::endl;
    exit(1);
  }
  return program;
}

// OpenCL kernel
cl_kernel createKernel(cl_program program, const char* kernelName){
  cl_kernel kernel;
  cl_int err;
  kernel = clCreateKernel(program, kernelName, &err);
  if (!kernel || err != CL_SUCCESS) {
    std::cerr << "Error: Failed to create compute kernel!" << std::endl;
    exit(1);
  }
  return kernel;
}
//*********************************//
//       inicializacio vege        //
//*********************************//
#pragma endregion R_CL_INIT


//*********************************//
//       PET Geometria leiras      //
//*********************************//
#pragma region R_GEOMETRY_DESC
struct module
{
    cl_float4 origin;
    cl_float4 axial;
    cl_float4 transAxial;
    cl_float4 n;
};

const int nModules = 4;
const int coincidence = 1;
const int nPairs = nModules*coincidence / 2;

const int nAxial = 32;
const int nTransAxial = 32;
const int lorNumberPerPair = nAxial*nTransAxial*nAxial*nTransAxial;
const int lorNumber = lorNumberPerPair*nPairs;
const float detectorArea = 2.0f*2.0f / nAxial / nTransAxial;
cl_float4 volumeMin;
cl_float4 volumeMax;
#pragma endregion R_GEOMETRY_DESC

//***********************//
//       PET adatok      //
//***********************//
#pragma region R_PET_DATA
cl_mem estimatedLorBufferGPU;
cl_mem measuredLorBufferGPU;
float estimatedLorBufferCPU[lorNumber];
float measuredLorBufferCPU[lorNumber];

cl_mem parallelProjectionBufferGPU;
float parallelProjectionBufferCPU[nPairs*nAxial*nTransAxial];

cl_mem moduleBufferGPU;
module moduleBufferCPU[nModules];

cl_mem volumeBufferGPU;
float* volumeData;

const int resolution = 32;
#pragma endregion R_PET_DATA

//*************************//
//       IO muveletek      //
//     (kep mentes, stb.)  //
//*************************//
#pragma region R_IO
void saveTGA(int pictureNumber, int width, int height, float* buffer){
  char* fname = (char*)malloc(sizeof(char)*32);
  sprintf(fname, "out_%03d.tga", pictureNumber);
  FILE *f = fopen(fname, "wb");
  if(!f) {
    fprintf(stderr, "Unable to create output targa image `%s'\n", "out.tga");
    exit(EXIT_FAILURE);
  }

  fputc(0x00, f); /* ID Length, 0 => No ID        */
  fputc(0x00, f); /* Color Map Type, 0 => No color map included   */
  fputc(0x02, f); /* Image Type, 2 => Uncompressed, True-color Image */
  fputc(0x00, f); /* Next five bytes are about the color map entries */
  fputc(0x00, f); /* 2 bytes Index, 2 bytes length, 1 byte size */
  fputc(0x00, f);
  fputc(0x00, f);
  fputc(0x00, f);
  fputc(0x00, f); /* X-origin of Image    */
  fputc(0x00, f);
  fputc(0x00, f); /* Y-origin of Image    */
  fputc(0x00, f);
  fputc(width & 0xff, f); /* Image Width      */
  fputc((width >> 8) & 0xff, f);
  fputc(height & 0xff, f); /* Image Height     */
  fputc((height >> 8) & 0xff, f);
  fputc(0x18, f); /* Pixel Depth, 0x18 => 24 Bits */
  fputc(0x20, f); /* Image Descriptor     */

  float vmax = 0;
  for(int y = 0; y < height; ++y) {
    for(size_t x = 0 ; x < width; ++x) {
        vmax = std::max(buffer[x + y * width],vmax);
    }
  }
  for(int y = 0; y < height; ++y) {
    for(size_t x = 0 ; x < width; ++x) {
        //buffer[x + y * width] *= (height*width) / vmax;
        buffer[x + y * width] /= vmax;
    }
  }


  for(int y = 0; y < height; ++y) {
    for(size_t x = 0 ; x < width; ++x) {
      fputc((int)(buffer[x + y * width] * 255.0f), f); /* write blue */
      fputc((int)(buffer[x + y * width] * 255.0f), f); /* write green */
      fputc((int)(buffer[x + y * width] * 255.0f), f); /* write red */
    }
  }
  fclose(f);
  free(fname);
}

void saveLorTGA(unsigned int iterationCnt){
    CL_SAFE_CALL( clEnqueueReadBuffer(commands, estimatedLorBufferGPU, CL_TRUE, 0, sizeof(float) * nPairs*nAxial*nTransAxial*nAxial*nTransAxial,
				    estimatedLorBufferCPU, 0, NULL, NULL));
    saveTGA(100+iterationCnt,nAxial*nTransAxial,nTransAxial*nAxial*nPairs,estimatedLorBufferCPU);
}

std::ostream& operator<< ( std::ostream& os, cl_float4& v )
{
   os << "(" << v.s[0] << "," << v.s[1] << "," << v.s[2] << "," << v.s[3] << ")";
   return os;
}
#pragma endregion R_IO

//*******************************************************//
//       PET rekon kernelek, algoritmus beallitasok      //
//*******************************************************//
#pragma region R_ALGORITHM
cl_program petProgram;

cl_kernel forwardProjectionKernel;
cl_kernel backProjectionKernel;
cl_kernel parallelProjectionKernel;

int iteration = 0;
int measurementGeneration = 0;
#pragma endregion R_ALGORITHM

//***************************************************************//
//       Volume vizualizacios kernelek, adatok, parameterek      //
//***************************************************************//
#pragma region R_VISDATA
cl_kernel visualizationKernel;

int width = 600;
int height = 600;
cl_mem visualizationBufferGPU;
cl_float4* visualizationBufferCPU;
Camera camera;
#pragma endregion R_VISDATA


//**************************************//
//       PET rekon inicializacioja      //
//**************************************//
#pragma region R_PETINIT
void rotate90xy( cl_float4 &vector )
{
    std::swap( vector.s[0], vector.s[1] );
    vector.s[0] *= -1.0f;
}

void initReconstruction(){
  petProgram = createProgram(context, device_id, "programs.cl");
  forwardProjectionKernel = createKernel(petProgram, "forwardProjection");
  backProjectionKernel = createKernel(petProgram, "backProjection");
  parallelProjectionKernel = createKernel(petProgram, "parallelProjection");
  visualizationKernel = createKernel(petProgram, "visualization");

  // lor tomb
  estimatedLorBufferGPU = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float) * lorNumber, NULL, NULL);
  measuredLorBufferGPU = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float) * lorNumber, NULL, NULL);

  // modul tomb
  cl_float4 origin, axial, transAxial, n;
  origin.s[3] = axial.s[3] = transAxial.s[3] = n.s[3] = 0.0f;
  origin.s[0] = origin.s[1] = origin.s[2] = -1.0f;                      // (-1,-1,-1)
  axial.s[2] = 2.0f; axial.s[0] = axial.s[1] = 0.0f;                    // (0,0,2)
  transAxial.s[0] = 2.0f; transAxial.s[1] = transAxial.s[2] = 0.0f;     // (2,0,0)
  n.s[1] = 1.0f; n.s[0] = n.s[2] = 0.0f;                                // (0,1,0)
  // modul tomb
  moduleBufferGPU = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(module)*nModules, NULL, NULL);
  for ( int i = 0; i < nModules; ++i )
  {
    moduleBufferCPU[i].origin = origin;
    moduleBufferCPU[i].axial = axial;
    moduleBufferCPU[i].transAxial = transAxial;
    moduleBufferCPU[i].n = n;

    rotate90xy(origin);
    rotate90xy(axial);
    rotate90xy(transAxial);
    rotate90xy(n);
  }
  CL_SAFE_CALL( clEnqueueWriteBuffer(commands, moduleBufferGPU,
                                     CL_TRUE, 0, sizeof(module)*nModules,
                                     moduleBufferCPU, 0, NULL, NULL));

  for ( unsigned int i = 0; i < nModules; ++i )
  {
      std::cout << i << ": a=" << moduleBufferCPU[i].axial << ", t=" << moduleBufferCPU[i].transAxial << ", o=" << moduleBufferCPU[i].origin << ", n=" << moduleBufferCPU[i].n << std::endl;
  }

  volumeMin.s[0] = volumeMin.s[1] = volumeMin.s[2] = -0.5f;
  volumeMax.s[0] = volumeMax.s[1] = volumeMax.s[2] = 0.5f;

  volumeMin.s[3] = volumeMax.s[3] = 0.0f;

  volumeData = new float[resolution * resolution * resolution];
  for ( unsigned int x = 0; x < resolution; ++x )
   for ( unsigned int y = 0; y < resolution; ++y )
      for ( unsigned int z = 0; z < resolution; ++z ) {
         if ( (x-resolution/2)*(x-resolution/2)+(y-resolution/2)*(y-resolution/2)+(z-resolution/2)*(z-
               resolution/2) < 3*(resolution/5)*(resolution/5) )
            volumeData[ x + resolution*y + resolution*resolution*z ] = 1.0f;
         else
            volumeData[ x + resolution*y + resolution*resolution*z ] = 0.0f;
      }
  // volume adat
  volumeBufferGPU = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float) * resolution * resolution * resolution, NULL, NULL);
  CL_SAFE_CALL( clEnqueueWriteBuffer(commands, volumeBufferGPU,
                                     CL_TRUE, 0, sizeof(float) * resolution * resolution * resolution,
                                     volumeData, 0, NULL, NULL));
   // kezdeti elorevetites a "meres" eloallitasara
  measurementGeneration = 1;
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 0, sizeof(int), &nAxial) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 1, sizeof(int), &nTransAxial) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 2, sizeof(int), &resolution) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 3, sizeof(float), &detectorArea) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 4, sizeof(cl_float4), &volumeMin) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 5, sizeof(cl_float4), &volumeMax) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 6, sizeof(cl_mem), &moduleBufferGPU) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 7, sizeof(cl_mem), &measuredLorBufferGPU) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 8, sizeof(cl_mem), &estimatedLorBufferGPU) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 9, sizeof(cl_mem), &volumeBufferGPU) );
  CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 10, sizeof(int), &measurementGeneration) );
  size_t forwardProjectionSize[3];
  forwardProjectionSize[0] = nPairs;
  forwardProjectionSize[1] = forwardProjectionSize[2] = nTransAxial*nAxial;
  CL_SAFE_CALL( clEnqueueNDRangeKernel(commands, forwardProjectionKernel,
  				       3, NULL, forwardProjectionSize, NULL,
  				       0, NULL, NULL) );
  clFinish(commands);
  measurementGeneration = 0;

  volumeData = new float[resolution * resolution * resolution];
  for ( unsigned int x = 0; x < resolution; ++x )
    for ( unsigned int y = 0; y < resolution; ++y )
      for ( unsigned int z = 0; z < resolution; ++z )
      {
        volumeData[ x +resolution*y +resolution*resolution*z ] = 1.0f;
      }

  // volume buffer
  CL_SAFE_CALL( clEnqueueWriteBuffer(commands, volumeBufferGPU,
                                     CL_TRUE, 0, sizeof(float) * resolution * resolution * resolution,
                                     volumeData, 0, NULL, NULL));

  // parhuzamos vetites buffere
  parallelProjectionBufferGPU = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float) * nPairs*nAxial*nTransAxial, NULL, NULL);
}
#pragma endregion R_PETINIT

//**********************************************//
//       PET rekon egy iteraciojanak kodja      //
//**********************************************//
#pragma region R_PET_RECONSTRUCTION
void reconstructionStep(){

    std::cout << "Computing forward projection...";

   // forward projection
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 0, sizeof(int), &nAxial) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 1, sizeof(int), &nTransAxial) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 2, sizeof(int), &resolution) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 3, sizeof(float), &detectorArea) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 4, sizeof(cl_float4), &volumeMin) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 5, sizeof(cl_float4), &volumeMax) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 6, sizeof(cl_mem), &moduleBufferGPU) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 7, sizeof(cl_mem), &measuredLorBufferGPU) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 8, sizeof(cl_mem), &estimatedLorBufferGPU) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 9, sizeof(cl_mem), &volumeBufferGPU) );
   CL_SAFE_CALL( clSetKernelArg(forwardProjectionKernel, 10, sizeof(int), &measurementGeneration) );

   size_t forwardProjectionSize[3];
   forwardProjectionSize[0] = nPairs;
   forwardProjectionSize[1] = forwardProjectionSize[2] = nTransAxial*nAxial;
   CL_SAFE_CALL( clEnqueueNDRangeKernel(commands, forwardProjectionKernel,
   				       3, NULL, forwardProjectionSize, NULL,
   				       0, NULL, NULL) );
   clFinish(commands);

   std::cout << "Done." << std::endl << "Computing Back projection...";

   // backProjection
   CL_SAFE_CALL( clSetKernelArg(backProjectionKernel, 0, sizeof(int), &nPairs) );
   CL_SAFE_CALL( clSetKernelArg(backProjectionKernel, 1, sizeof(int), &nAxial) );
   CL_SAFE_CALL( clSetKernelArg(backProjectionKernel, 2, sizeof(int), &nTransAxial) );
   CL_SAFE_CALL( clSetKernelArg(backProjectionKernel, 3, sizeof(int), &resolution) );
   CL_SAFE_CALL( clSetKernelArg(backProjectionKernel, 4, sizeof(cl_float4), &volumeMin) );
   CL_SAFE_CALL( clSetKernelArg(backProjectionKernel, 5, sizeof(cl_float4), &volumeMax) );
   CL_SAFE_CALL( clSetKernelArg(backProjectionKernel, 6, sizeof(cl_mem), &moduleBufferGPU) );
   CL_SAFE_CALL( clSetKernelArg(backProjectionKernel, 7, sizeof(cl_mem), &estimatedLorBufferGPU) );
   CL_SAFE_CALL( clSetKernelArg(backProjectionKernel, 8, sizeof(cl_mem), &volumeBufferGPU) );

   size_t backProjectionSize[3];
   backProjectionSize[0] = backProjectionSize[1] = backProjectionSize[2] = resolution;
   CL_SAFE_CALL( clEnqueueNDRangeKernel(commands, backProjectionKernel,
   				       3, NULL, backProjectionSize, NULL,
   				       0, NULL, NULL) );
   clFinish(commands);

   std::cout << "Done." << std::endl << "Computing parallel projection...";

   // parallel
   CL_SAFE_CALL( clSetKernelArg(parallelProjectionKernel, 0, sizeof(int), &nAxial) );
   CL_SAFE_CALL( clSetKernelArg(parallelProjectionKernel, 1, sizeof(int), &nTransAxial) );
   CL_SAFE_CALL( clSetKernelArg(parallelProjectionKernel, 2, sizeof(cl_mem), &estimatedLorBufferGPU) );
   CL_SAFE_CALL( clSetKernelArg(parallelProjectionKernel, 3, sizeof(cl_mem), &parallelProjectionBufferGPU) );

   size_t parallelProjectionSize[3];
   parallelProjectionSize[0] = nPairs;
   parallelProjectionSize[1] = nAxial;
   parallelProjectionSize[2] = nTransAxial;
   CL_SAFE_CALL( clEnqueueNDRangeKernel(commands, parallelProjectionKernel,
   				       3, NULL, parallelProjectionSize, NULL,
   				       0, NULL, NULL) );
   clFinish(commands);

   std::cout << "Done." << std::endl;

   // save parallel projection
   CL_SAFE_CALL( clEnqueueReadBuffer(commands, parallelProjectionBufferGPU, CL_TRUE, 0, sizeof(float) * nPairs*nAxial*nTransAxial,
				    parallelProjectionBufferCPU, 0, NULL, NULL));

   saveTGA(iteration,nAxial,nTransAxial*nPairs,parallelProjectionBufferCPU);
}
#pragma endregion R_PET_RECONSTRUCTION


//*****************************************************************************//
//                 Megjelenito kod, az elmult gyakorlatok alapjan              //
//                                                                             //
//       Nem resze a rekonstrukcionak, a kiszamolt eredmenyt jeleniti meg      //
//*****************************************************************************//
#pragma region R_VISUALIZATION
void resetSimulation(int resolution, cl_mem volumeBufferGPU){
  cl_kernel resetSimulationKernel = createKernel(petProgram, "resetSimulation");

  CL_SAFE_CALL(clSetKernelArg(resetSimulationKernel, 0, sizeof(int), &resolution));
  CL_SAFE_CALL(clSetKernelArg(resetSimulationKernel, 1, sizeof(cl_mem), &volumeBufferGPU));

  size_t workSize[3];
  workSize[0] = resolution;
  workSize[1] = resolution;
  workSize[2] = resolution;

  CL_SAFE_CALL(clEnqueueNDRangeKernel(commands, resetSimulationKernel,
				      3, NULL, workSize, NULL,
				      0, NULL, NULL));
  clFinish(commands);

  clReleaseKernel(resetSimulationKernel);
}

void visualizationStep(){
  cl_float16 clViewDir;
  float* camMatrix = camera.getViewDirMatrix().getPointer();
  for(int i = 0; i < 16; ++i){
    clViewDir.s[i] = camMatrix[i];
  }

  Vector eye = camera.getEye();
  clViewDir.s[12] = eye.x;
  clViewDir.s[13] = eye.y;
  clViewDir.s[14] = eye.z;
  clViewDir.s[15] = 1.0f;

  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 0, sizeof(int), &width) );
  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 1, sizeof(int), &height) );
  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 2, sizeof(cl_mem), &visualizationBufferGPU) );
  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 3, sizeof(int), &resolution) );
  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 4, sizeof(cl_mem), &volumeBufferGPU) );
  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 5, sizeof(int), &iteration) );
  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 6, sizeof(cl_float16), &clViewDir));

  size_t visualizationBufferSize[2];
  visualizationBufferSize[0] = width;
  visualizationBufferSize[1] = height;

  CL_SAFE_CALL(clEnqueueNDRangeKernel(commands, visualizationKernel,
				      2, NULL, visualizationBufferSize, NULL,
				      0, NULL, NULL) );

  clFinish(commands);

  CL_SAFE_CALL( clEnqueueReadBuffer(commands, visualizationBufferGPU, CL_TRUE, 0, sizeof(cl_float4) * width * height,
				    visualizationBufferCPU, 0, NULL, NULL));

  //saveTGA(iteration, width, height, visualizationBufferCPU);
  glDrawPixels(width, height, GL_RGBA, GL_FLOAT, visualizationBufferCPU);
}

// Common data
void initCommon(){
  // visualization buffer
  visualizationBufferCPU = new cl_float4[width*height];
  visualizationBufferGPU = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(cl_float4) * width * height, NULL, NULL);
}

// OpenGL
int method = 1;
bool keysPressed[256];

void initOpenGL(){
  glewExperimental = GL_TRUE;
  GLenum err = glewInit();
  if (GLEW_OK != err)     {
    std::cerr << "Error: "<< glewGetErrorString(err) << std::endl;
  } else {
    if (GLEW_VERSION_3_0)
      {
	std::cout<< "Driver supports OpenGL 3.0\nDetails:"<< std::endl;
	std::cout << "  Using GLEW " << glewGetString(GLEW_VERSION)<< std::endl;
	std::cout << "  Vendor: " << glGetString (GL_VENDOR)<< std::endl;
	std::cout << "  Renderer: " << glGetString (GL_RENDERER)<< std::endl;
	std::cout << "  Version: " << glGetString (GL_VERSION)<< std::endl;
	std::cout << "  GLSL: " << glGetString (GL_SHADING_LANGUAGE_VERSION)<< std::endl;
      }
  }

  glClearColor(0.17f, 0.4f, 0.6f, 1.0f);
}

void display(){
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glDisable(GL_DEPTH_TEST);

  visualizationStep();

  glEnable(GL_DEPTH_TEST);
  glutSwapBuffers();
}

void animate(){
        static float lastTime = 0.0;
        long timeInMilliSecs = glutGet(GLUT_ELAPSED_TIME);
        float timeNow = timeInMilliSecs / 1000.0f;
        float deltaTime = timeNow - lastTime;
        lastTime = timeNow;

        camera.control(deltaTime, keysPressed);

	glutPostRedisplay();
}

void keyDown(unsigned char key, int x, int y){
  keysPressed[key] = true;
}

void keyUp(unsigned char key, int x, int y){
  keysPressed[key] = false;
  switch(key){
  case 'i':
  case 'I':
      ++iteration;
      reconstructionStep();
      break;
  case 'l':
  case 'L':
      saveLorTGA(iteration);
      break;
  case 27:
    exit(0);
    break;
  }
}

void mouseClick(int button, int state, int x, int y){
  if(button == GLUT_LEFT_BUTTON)
    if(state == GLUT_DOWN)
      camera.startDrag(x, y);
}

void mouseMove(int x, int y){
  camera.drag(x, y);
}

void reshape(int newWidth, int newHeight){
  width = newWidth;
  height = newHeight;
  glViewport(0, 0, width, height);
  camera.setAspectRatio((float)width/height);
}
#pragma endregion R_VISUALIZATION


//********************************//
//       main fuggveny, glut      //
//********************************//
int main(int argc, char* argv[]){
  // OpenCL init
  platform = createPlatform();
  device_id = createDevice(platform, CL_DEVICE_TYPE_GPU);
  context = createContext(device_id);
  commands = createCommandQueue(context, device_id);


  glutInit(&argc, argv);
  glutInitContextVersion (3, 0);
  glutInitContextFlags (GLUT_CORE_PROFILE | GLUT_DEBUG);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
  glutInitWindowSize(width, height);
  glutCreateWindow("GPGPU 12. labor: PETi");

  initOpenGL();

  glutDisplayFunc(display);
  glutIdleFunc(animate);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyDown);
  glutKeyboardUpFunc(keyUp);
  glutMouseFunc(mouseClick);
  glutMotionFunc(mouseMove);

  // OpenCL, rekonstrukcio inicializacio
  initCommon();
  initReconstruction();

  glutMainLoop();

  // OpenCL cleanup
  clReleaseProgram(petProgram);
  clReleaseCommandQueue(commands);
  clReleaseContext(context);
  return(0);
}
